try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
from ._widget import DeepmetaDemoWidget, DeepmetaWidget

__all__ = ("DeepmetaWidget", "DeepmetaDemoWidget")



import os

import cc3d
import cv2
import numpy as np
import skimage.exposure as exposure
import skimage.measure as measure
import skimage.transform as transform
import torch
from scipy import ndimage

import napari_deepmeta.models as models

################
# SEGMENTATION #
################


def process_img(mouse, model):
    output_stack = []
    for slice in mouse:
        slice = slice.reshape(1, 1, 128, 128)
        slice = torch.from_numpy(slice).float()
        output = model(slice)
        output = output.max(1).indices
        output_stack.append(output.cpu().detach().numpy())
    return output_stack


def contrast_and_reshape(mouse, size=128):
    # if len(mouse.shape) > 2:
    data = []
    for i in np.arange(mouse.shape[0]):
        img_adapteq = exposure.equalize_adapthist(mouse[i], clip_limit=0.03)
        data.append(img_adapteq)
    return np.array(data).reshape(-1, 1, size, size)
    # else:
    #     img_adapteq = exposure.equalize_adapthist(mouse, clip_limit=0.03)
    #     return np.array(img_adapteq).reshape(1, size, size)


def segment_stack(img):
    device = "cuda" if torch.cuda.is_available() else "cpu"  # marche pas
    model = models.Unet3plus()  # .to(device)
    model.load_state_dict(
        torch.load(
            os.path.dirname(os.path.realpath(__file__))
            + "/resources/model.pth",
            map_location=torch.device(device),
        )
    )
    model.eval()
    return model(img)  # .to(device))


def prepare_mouse(img, contrast):
    img = transform.resize(img, (len(img), 128, 128), anti_aliasing=True)
    img = img / np.amax(img)
    if contrast:
        return torch.Tensor(contrast_and_reshape(img))
    else:
        return torch.Tensor(img.reshape(-1, 1, 128, 128))


##################
# POSTPROCESSING #
##################


def remove_blobs(mask, min_size=10):
    mask = mask.reshape(128, 128).astype(np.uint8)
    nb_components, output, stats, centroids = cv2.connectedComponentsWithStats(
        mask, connectivity=8
    )
    sizes = stats[1:, -1]
    nb_components = nb_components - 1  # remove background
    img2 = np.zeros(output.shape)
    # for every component in the image, you keep it only if it's above min_size
    for i in range(nb_components):
        if sizes[i] >= min_size:
            img2[output == i + 1] = 255
    return img2


def dilate_and_erode(img, k1=3, k2=3):
    kernel1 = np.ones((k1, k1), np.uint8)
    kernel2 = np.ones((k2, k2), np.uint8)
    img_dilation = cv2.dilate(img, kernel1, iterations=1)
    return cv2.erode(img_dilation, kernel2, iterations=1)


def laplace(img_stack, mask_list):
    img_stack2 = (img_stack * 255).astype(np.uint8)
    for i, img in enumerate(img_stack2):
        new_im = ndimage.gaussian_laplace(img, sigma=7)
        if np.mean(new_im) < 1:
            mask_list[i] = np.zeros((128, 128))
        else:
            break
    for i, img in enumerate(img_stack2[::-1]):
        new_im = ndimage.gaussian_laplace(img, sigma=7)
        if np.mean(new_im) < 1:
            mask_list[(len(mask_list) - 1) - i] = np.zeros((128, 128))
        else:
            break
    return np.array(mask_list)


def sanity_check(mask_list):
    mask_list[0] = np.zeros((128, 128))
    mask_list[-1] = np.zeros((128, 128))
    for i in range(1, len(mask_list) - 1):
        if mask_list[i].sum() > 15:
            if mask_list[i - 1].sum() < 15 and mask_list[i + 1].sum() < 15:
                mask_list[i] = np.zeros((128, 128))
        else:
            mask_list[i] = np.zeros((128, 128))
    return mask_list


def postprocess(inputs, masks):
    inputs = inputs.detach().numpy()
    masks = laplace(inputs, masks)
    lungs_masks = np.array([mask > 0.5 for mask in masks])
    metas_masks = np.array([mask > 1.5 for mask in masks])
    lungs_masks = sanity_check(lungs_masks)
    lungs_masks = [remove_blobs(mask, 10) for mask in lungs_masks]
    lungs_masks = (
        np.array([dilate_and_erode(mask, 3, 3) for mask in lungs_masks]) / 255
    )
    metas_masks = [remove_blobs(mask, 3) for mask in metas_masks]
    metas_masks = (
        np.array([dilate_and_erode(mask, 3, 3) for mask in metas_masks]) / 255
    )
    return np.where((metas_masks == 1), 2, lungs_masks)


###########
# Drawing #
###########


def add_z(arr, z):
    return [np.insert(couple, 0, z) for couple in arr]


def border_detected(k, seg):
    return measure.find_contours(seg[k], 0.8)


def mask_to_plottable_3D(masks):
    plottable_list = []
    for i in range(len(masks)):
        contours = border_detected(i, masks)
        for contour in contours:
            contour = add_z(contour, i)
            if len(contour) > 0:
                plottable_list.append(contour)
    return plottable_list


def get_meta_nb(masks):
    _, N = cc3d.connected_components(
        np.array(masks), return_N=True, connectivity=18
    )
    return N



import numpy as np

# make_napari_viewer is a pytest fixture that returns a napari viewer object
# capsys is a pytest fixture that captures stdout and stderr output streams
# def test_deepmeta_widget(make_napari_viewer, capsys):
#     # make viewer and add an image layer using our fixture
#     viewer = make_napari_viewer()
#     viewer.add_image(np.random.random((100, 100)))
#
#     # create our widget, passing in the viewer
#     my_widget = DeepmetaWidget(viewer)
#
#     # call our widget method
#     my_widget._on_click()
#
#     # read captured output and check that it's as we expected
#     captured = capsys.readouterr()
#     assert captured.out == "done\n"


#
# def test_demo_widget(make_napari_viewer, capsys):
#     # make viewer and add an image layer using our fixture
#     viewer = make_napari_viewer()
#     viewer.add_image(np.random.random((100, 100)))
#
#     # create our widget, passing in the viewer
#     my_widget = DeepmetaDemoWidget(viewer)
#
#     # call our widget method
#     my_widget._on_click()
#
#     # read captured output and check that it's as we expected
#     captured = capsys.readouterr()
#     assert captured.out == "done\n"


def test_add_z():
    import napari_deepmeta.deepmeta_functions as df

    arr = np.array([[0, 1], [0, 1]])
    assert (df.add_z(arr, 3) == np.array([[3, 0, 1], [3, 0, 1]])).all()


def test_dilate_and_erode():
    import napari_deepmeta.deepmeta_functions as df

    img = np.zeros((128, 128))
    res = df.dilate_and_erode(img)
    assert np.shape(img) == np.shape(res)


def test_contrast():
    import napari_deepmeta.deepmeta_functions as df

    seg = np.ones((2, 128, 128))
    res = df.contrast_and_reshape(seg)
    assert np.shape(res) == (2, 1, 128, 128)



import torch
import torch.nn as nn


class SeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        bias=False,
        padding="same",
        stride=1,
    ):
        super().__init__()
        padding = 1 if stride > 1 else "same"
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            groups=in_channels,
            bias=bias,
            padding=padding,
            stride=stride,
        )
        self.pointwise = nn.Conv2d(
            in_channels, out_channels, kernel_size=1, bias=bias, padding="same"
        )

    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out


class DoubleConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        drop_r,
        conv=SeparableConv2d,
    ):
        super().__init__()
        self.double_conv = nn.Sequential(
            conv(in_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
            nn.Dropout2d(drop_r),
            conv(out_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(),
            nn.Dropout2d(drop_r),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down_Block(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        drop=0.1,
        conv_l=SeparableConv2d,
    ):
        super().__init__()
        self.conv = DoubleConv(in_channels, out_channels, drop, conv=conv_l)
        self.down = nn.MaxPool2d(2)

    def forward(self, x):
        c = self.conv(x)
        return c, self.down(c)


class Bridge(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        drop=0.1,
        conv_l=SeparableConv2d,
    ):
        super().__init__()
        self.conv = DoubleConv(in_channels, out_channels, drop, conv=conv_l)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels, conv_l=SeparableConv2d):
        super().__init__()
        self.conv = conv_l(in_channels, out_channels, kernel_size=1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv(x)


class Normalize_Down(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel: int):
        super().__init__()
        self.down = nn.MaxPool2d(kernel)
        self.norm = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, padding="same"
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(self.down(x))


class Normalize_Up(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel: int):
        super().__init__()
        self.up = nn.UpsamplingBilinear2d(scale_factor=kernel)
        self.norm = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, padding="same"
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(self.up(x))


class Concat_Block(nn.Module):
    def __init__(
        self,
        kernels_down,
        filters_down,
        kernels_up,
        filters_up,
    ):
        super().__init__()
        self.norm_down = nn.ModuleList(
            [
                Normalize_Down(in_, 32, kernel)
                for (kernel, in_) in zip(kernels_down, filters_down)
            ]
        )
        self.norm_up = nn.ModuleList(
            [
                Normalize_Up(in_, 32, kernel)
                for (kernel, in_) in zip(kernels_up, filters_up)
            ]
        )

    def forward(self, down, up):
        res = [l(d) for d, l in zip(down, self.norm_down)]
        res.extend(l(u) for u, l in zip(up, self.norm_up))
        return torch.cat(res, dim=1)


class Up_Block_3p(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels: int,
        drop=0.1,
        conv_l=SeparableConv2d,
    ):
        super().__init__()
        self.up = nn.ConvTranspose2d(
            in_channels, 32, kernel_size=(2, 2), stride=(2, 2)
        )
        self.conv = DoubleConv(160, 160, drop, conv=conv_l)

    def forward(self, x: torch.Tensor, conc: torch.Tensor) -> torch.Tensor:
        x1 = self.up(x)
        x = torch.cat([conc, x1], dim=1)
        return self.conv(x)


class Unet3plus(nn.Module):
    def __init__(
        self,
        filters=16,
        classes=3,
        drop_r=0.1,
        conv_l=SeparableConv2d,
    ):
        """
        Initialize the U-Net model.
        :param filters: Number of filters in the first convolutional layer.
        :type filters: int
        :param classes: Number of classes to predict.
        :type classes: int
        :param drop_r: Dropout rate.
        :type drop_r: float
        """
        super().__init__()

        self.down1 = Down_Block(1, filters, conv_l=conv_l)
        self.down2 = Down_Block(filters, filters * 2, drop_r, conv_l=conv_l)
        self.down3 = Down_Block(
            filters * 2, filters * 4, drop_r, conv_l=conv_l
        )
        self.down4 = Down_Block(
            filters * 4, filters * 8, drop_r, conv_l=conv_l
        )

        self.bridge = Bridge(filters * 8, filters * 16, drop_r, conv_l=conv_l)

        self.up1 = Up_Block_3p(filters * 16, 160, drop_r, conv_l=conv_l)
        self.up2 = Up_Block_3p(160, 160, drop_r, conv_l=conv_l)
        self.up3 = Up_Block_3p(160, 160, drop_r, conv_l=conv_l)
        self.up4 = Up_Block_3p(160, 160, drop_r, conv_l=conv_l)

        self.concat1 = Concat_Block(
            kernels_down=[8, 4, 2, 1],
            filters_down=[filters, filters * 2, filters * 4, filters * 8],
            kernels_up=[],
            filters_up=[],
        )
        self.concat2 = Concat_Block(
            kernels_down=[4, 2, 1],
            filters_down=[filters, filters * 2, filters * 4],
            kernels_up=[4],
            filters_up=[filters * 16],
        )
        self.concat3 = Concat_Block(
            kernels_down=[2, 1],
            filters_down=[filters, filters * 2],
            kernels_up=[4, 8],
            filters_up=[160, filters * 16],
        )
        self.concat4 = Concat_Block(
            kernels_down=[1],
            filters_down=[filters],
            kernels_up=[4, 8, 16],
            filters_up=[160, 160, filters * 16],
        )
        self.outc = OutConv(160, classes, conv_l=conv_l)

    def forward(self, x):
        c1, x1 = self.down1(x)
        c2, x2 = self.down2(x1)
        c3, x3 = self.down3(x2)
        c4, x4 = self.down4(x3)
        bridge = self.bridge(x4)
        x5 = self.up1(bridge, self.concat1([c1, c2, c3, c4], []))
        x6 = self.up2(x5, self.concat2([c1, c2, c3], [bridge]))
        x7 = self.up3(x6, self.concat3([c1, c2], [x5, bridge]))
        x8 = self.up4(x7, self.concat4([c1], [x6, x5, bridge]))
        return self.outc(x8)



from typing import TYPE_CHECKING

import numpy as np
from qtpy import QtCore
from qtpy.QtWidgets import (
    QCheckBox,
    QLabel,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

if TYPE_CHECKING:
    pass


def load_img(obj):
    import os

    import skimage.io as io

    import napari_deepmeta.deepmeta_functions as df

    img = io.imread(
        os.path.dirname(os.path.realpath(__file__))
        + "/resources/souris_8.tif",
        plugin="tifffile",
    )
    img = df.contrast_and_reshape(img).reshape(128, 128, 128)
    obj.viewer.add_image(img, name="mouse")
    return img


def show_total_vol(layout, masks, organ, nb=None):
    vol_tot = masks.sum() * 0.0047
    elt = QLabel(f"Total volume {organ} {vol_tot:.3f}mm3")
    layout.addWidget(elt)
    if nb is not None:
        elt2 = QLabel(f"Metastases number {nb}")
        layout.addWidget(elt2)


def clean_labels(layout):
    if layout.count() >= 6:
        try:
            layout.takeAt(5).widget().setParent(None)
            layout.takeAt(5).widget().setParent(None)
            layout.takeAt(5).widget().setParent(None)
        except AttributeError:
            pass


def show_error(text):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Critical)
    msg.setText("Error")
    msg.setInformativeText(text)
    msg.setWindowTitle("Error")
    msg.exec_()


def show_shapes_3D(viewer, plottable, color, text):
    try:
        viewer.add_shapes(
            plottable,
            shape_type="path",
            edge_width=0.5,
            edge_color=color,
            face_color="#6a6a6aff",
            opacity=0.6,
            name=text,
        )
    except Exception as e:
        print(e)
        print(np.shape(plottable))


class DeepmetaWidget(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer

        btn = QPushButton("Segment Stack")
        btn.clicked.connect(self._on_click)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(btn)

        check = QCheckBox("Contrast ?", self)
        check.stateChanged.connect(self._click_box)
        self.contrast = False
        self.layout().addWidget(check)

        check2 = QCheckBox("Post-processing ?", self)
        check2.stateChanged.connect(self._click_box2)
        self.postprocess = False
        self.layout().addWidget(check2)

        check3 = QCheckBox("Show metastases ?", self)
        check3.stateChanged.connect(self._click_box3)
        self.metas = False
        self.layout().addWidget(check3)

    def _click_box(self, state):
        self.contrast = state == QtCore.Qt.Checked

    def _click_box2(self, state):
        self.postprocess = state == QtCore.Qt.Checked

    def _click_box3(self, state):
        self.metas = state == QtCore.Qt.Checked

    def _on_click(self):
        import napari_deepmeta.deepmeta_functions as df

        if len(self.viewer.layers) == 1:
            img = df.prepare_mouse(self.viewer.layers[0].data, self.contrast)
            output = df.segment_stack(img)
            output = output.max(1).indices.detach().numpy()
            if self.postprocess:
                output = df.postprocess(img, output)
            masks = [mask > 0.5 for mask in output]
            plottable_list = df.mask_to_plottable_3D(masks)
            clean_labels(self.layout())
            show_shapes_3D(self.viewer, plottable_list, "red", "Lung masks")
            show_total_vol(self.layout(), np.array(masks), "lungs")
            if self.metas:
                masks = [mask > 1.5 for mask in output]
                plottable_list = df.mask_to_plottable_3D(masks)
                show_shapes_3D(
                    self.viewer, plottable_list, "blue", "Metastases masks"
                )
                show_total_vol(
                    self.layout(),
                    np.array(masks),
                    "metastases",
                    nb=df.get_meta_nb(masks),
                )
            print("done")
        else:
            show_error(
                "Cannot run segmentation if you have multiple files opened."
            )


class DeepmetaDemoWidget(QWidget):
    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer

        btn = QPushButton("Run demo")
        btn.clicked.connect(self._on_click)

        self.setLayout(QVBoxLayout())
        self.layout().addWidget(btn)

    def _on_click(self):
        import napari_deepmeta.deepmeta_functions as df

        img = load_img(self)
        img = df.prepare_mouse(img, False)
        output = df.segment_stack(img)
        output = output.max(1).indices.detach().numpy()
        output = df.postprocess(img, output)
        masks = [mask > 0.5 for mask in output]
        plottable_list = df.mask_to_plottable_3D(masks)
        clean_labels(self.layout())
        show_shapes_3D(self.viewer, plottable_list, "red", "Lung masks")
        show_total_vol(self.layout(), np.array(masks), "lungs")
        masks = [mask > 1.5 for mask in output]
        plottable_list = df.mask_to_plottable_3D(masks)
        show_shapes_3D(self.viewer, plottable_list, "blue", "Metastases masks")
        show_total_vol(
            self.layout(),
            np.array(masks),
            "metastases",
            nb=df.get_meta_nb(masks),
        )
        print("done")
